/// \file 
/// \brief Definitions related to the AVL class

#ifndef AVL_H
#define AVL_H

#include "avl_node.h"
#include "variable.h"
#include <vector>
#include <light_set.h>
#include "avl_iterator.h"

class AVL {                    /// This class allows to represent AVL trees
	protected:	
		AVL_Node * root;					///< the root of the AVL
		AVL_Node *** tuple_list;	///< the list of tuples for each pair (variable,value) 
		int arity; 								///< the arity of the tuple in the tree
		unsigned long size;				///< the size of the AVL
	
	public:
		// constructors and destructor
		AVL (vector<Variable *> & var);		  	///< constructs an empty AVL for the tuples over the Cartesian products of the domains of the variables in var
		AVL (Light_Vector<Variable *> & var);	///< constructs an empty AVL for the tuples over the Cartesian products of the domains of the variables in var
		AVL (Light_Set<Variable *> & var);	  ///< constructs an empty AVL for the tuples over the Cartesian products of the domains of the variables in var
		~AVL ();													  	///< destructor

		// iterators
		typedef AVL_Iterator iterator;                  ///< defines the type representing iterators on AVL
		AVL::iterator begin (int var_pos, int val);			///< returns an iterator referring to the first element of the AVL for the variable at position var_pos in the associated scope and for the value val
		AVL::iterator begin (int var_pos, int * tuple);	///< returns an iterator referring to the element of the AVL containing tuple for the variable at position var_pos in the associated scope
		AVL::iterator end (int var_pos);					 			///< returns an iterator referring to the position after the last element of the AVL for the variable at position var_pos in the associated scope
		
		
		// basic functions
		bool Is_Empty ();								///< returns true if the AVL is empty, false otherwise
		unsigned long Get_Size ();			///< returns the size of the AVL (i.e. the number of elements it contains)
		void Add_Element (int * val);		///< adds the element val into the AVL (if it does not belong to the AVL yet)
		bool Is_Present (int * val);		///< returns true if the element val belongs to the AVL, false otherwise
		AVL_Node * Get_First_Tuple (int pos, int val);	///< returns the first node for the variable at position pos and for the val
};



//--------------------------------------
// inline iterator function definitions 
//--------------------------------------


// returns an iterator referring to the first element of the AVL for the variable at position var_pos in the associated scope and for the value val
inline AVL::iterator AVL::begin (int var_pos, int val) {
	return AVL_Iterator (this,var_pos,tuple_list[var_pos][val]);
}


// returns an iterator referring to the element of the AVL containing tuple for the variable at position var_pos in the associated scope
inline AVL::iterator AVL::begin (int var_pos, int * tuple) {
	return AVL_Iterator (this,var_pos,root->Get_Node (tuple,arity));
}


// returns an iterator referring to the position after the last element of the AVL for the variable at position var_pos in the associated scope
inline AVL::iterator AVL::end (int var_pos) {
	return AVL_Iterator (this,var_pos);
}


//-----------------------------------
// inline basic function definitions 
//-----------------------------------


// returns true if the AVL is empty, false otherwise
inline bool AVL::Is_Empty () {
	return root == 0;
}


// returns the size of the AVL (i.e. the number of elements it contains)
inline unsigned long AVL::Get_Size () {
	return size;
}	


// adds the element val into the AVL (if it does not belong to the AVL yet)
inline void AVL::Add_Element (int * val) {
	size++;
	if (root == 0)  root = new AVL_Node (val,arity,tuple_list);
	else root->Add_Element (root,val,arity,tuple_list);
}


// returns true if the element val belongs to the AVL, false otherwise
inline bool AVL::Is_Present (int * val) {
	if (root == 0) return 0;
	else return root->Is_Present (val,arity);
}


// returns the first node for the variable at position pos and for the val
inline AVL_Node * AVL::Get_First_Tuple (int pos, int val) {
	return tuple_list[pos][val];
}

#endif
