/// \file 
/// \brief Definitions related to the Constraint class

#ifndef CONSTRAINT_H
#define CONSTRAINT_H

#include <vector>
#include <set>

using namespace std;


class Assignment;
class CSP;

#include "multiple_residue_support.h"
#include "deletion_stack.h"
#include "tuple.h"

class Constraint {     /// This class allows to represent constraints
	protected:
		int num;																	///< the number of the constraint
		unsigned int arity;												///< the arity of the constraint
		set<unsigned int> scope;   								///< the number of the variables involved in the constraint
		Light_Set<Variable *> scope_var;					///< the variables involved in the constraint	
		map<int,int> position;										///< the position of a variable x in the scope of the constraint
		unsigned int unassigned_variable_number;	///< the number of unassigned variables in the scope of the constraint
		unsigned long weight;											///< the weight associated to the constraint as introduced in [Boussemart, Hemery, Lecoutre and Sais, 2004]
		
		
	public:
		// destructor
		Constraint (vector<Variable *> & var);			 ///< constructs a constraint whose scope is defined by the variables in var
		Constraint (Constraint & c); 								 ///< constructs a constraint by copying the constraint c
		virtual ~Constraint(){};			 							 ///< destructor
			
		// iterators
		set<unsigned int>::iterator Begin();   			 ///< returns an iterator referring to the first variable of the scope
    set<unsigned int>::iterator End();					 ///< returns an iterator referring to the position after the last variable of the scope
    
    Light_Set<Variable *>::iterator Begin_Var(); ///< returns an iterator referring to the first variable of the scope
    Light_Set<Variable *>::iterator End_Var();	 ///< returns an iterator referring to the position after the last variable of the scope
			
		// basic functions
		set<unsigned int> & Get_Scope ();	   				 				 ///< returns the scope of the constraint
		Light_Set<Variable *> & Get_Scope_Var ();		 				 ///< returns the scope of the constraint
		Variable * Get_Scope_Variable (int i);			 				 ///< returns the i th variable of the scope
		int Get_Position (int var);													 ///< returns the position of the variable var in the scope of the constraint, -1 if the variable does not belong to this scope
		unsigned int Get_Arity ();									 				 ///< returns the arity of the constraint
		unsigned long Get_Weight ();								 				 ///< returns the weight associated to the constraint
		void Set_Weight (unsigned long w);					 				 ///< sets the weight associated to the constraint to w
		void Inc_Weight ();								 					 				 ///< increases by one the weight associated to the constraint
		int Get_Num ();																			 ///< returns the number of the constraint
		void Set_Num (int n);																 ///< sets the number of the constraint to n
		bool Involve (int x);																 ///< returns true if the variable x is involved in the scope of the cosntraint, false otherwise	
		unsigned int Get_Unassigned_Variable_Number ();			 ///< returns the number of unassigned variables in the scope of the constraint
		void Inc_Unassigned_Variable_Number ();							 ///< increases by one the number of unassigned variables in the scope of the constraint
		void Dec_Unassigned_Variable_Number ();			 				 ///< decreases by one the number of unassigned variables in the scope of the constraint
		virtual bool Revise (CSP * pb, int var, Multiple_Residue_Support * ls, Deletion_Stack * ds)=0;	 ///< returns true if the application of arc-consistency on the constraint w.r.t. the variable var deletes a value in the domain of var, false otherwise
		virtual bool Has_Specific_Propagator ();		 				 ///< returns true if the constraint has a specific propagator, false otherwise
		virtual bool Propagate (CSP * pb, Multiple_Residue_Support * ls, Deletion_Stack * ds, bool * & is_modified);	 ///< returns true if the application of the propagator of the constraint deletes a value from the domain of at least one variable from the scope, false otherwise (is_modified[i] is true if the i th variable of the scope has been modified, false otherwise)
		virtual bool Propagate (CSP * pb, Assignment & A, Multiple_Residue_Support * ls, Deletion_Stack * ds, bool * & is_modified);	 ///< returns true if the application of the propagator of the constraint w.r.t. the assignment A deletes a value from the domain of at least one variable from the scope, false otherwise (is_modified[i] is true if the i th variable of the scope has been modified, false otherwise)
};


//-----------------------------
// inline function definitions
//-----------------------------


//-----------
// iterators
//-----------


// returns an iterator referring to the first variable of the scope
inline set<unsigned int>::iterator Constraint::Begin() {
	return scope.begin();
}



// returns an iterator referring to the position after the last variable of the scope
inline set<unsigned int>::iterator Constraint::End() {
	return scope.end();
}


// returns an iterator referring to the first variable of the scope
inline Light_Set<Variable *>::iterator Constraint::Begin_Var() {
	return scope_var.begin();
}


// returns an iterator referring to the position after the last variable of the scope
inline Light_Set<Variable *>::iterator Constraint::End_Var() {
	return scope_var.end();
}


//-----------------
// basic functions
//-----------------


// returns the scope of the constraint
inline set<unsigned int> & Constraint::Get_Scope () {
	return scope;
}


// returns the scope of the constraint
inline Light_Set<Variable *> & Constraint::Get_Scope_Var () {
	return scope_var;
}


// returns the i th variable of the scope
inline Variable * Constraint::Get_Scope_Variable (int i) {
	return scope_var[i];
}


// returns the arity of the constraint
inline unsigned int Constraint::Get_Arity () {
	return arity;
}


// returns the weight associated to the constraint
inline unsigned long Constraint::Get_Weight () {
	return weight;
}


// sets the weight associated to the constraint to w
inline void Constraint::Set_Weight (unsigned long w) {
	weight = w;
}


// increases by one the weight associated to the constraint
inline void Constraint::Inc_Weight () {
	weight++;
}


// returns the number of the constraint
inline int Constraint::Get_Num () {
	return num;
}


// sets the number of the constraint to n
inline void Constraint::Set_Num (int n) {
	num = n;
}


// returns true if the variable x is involved in the scope of the cosntraint, false otherwise
inline bool Constraint::Involve (int x) {
	return position.find (x) != position.end();
}


// returns the number of unassigned variables in the scope of the constraint
inline unsigned int Constraint::Get_Unassigned_Variable_Number () {
	return unassigned_variable_number;
}


// increases by one the number of unassigned variables in the scope of the constraint
inline void Constraint::Inc_Unassigned_Variable_Number () {
	unassigned_variable_number++;
}


// decreases by one the number of unassigned variables in the scope of the constraint
inline void Constraint::Dec_Unassigned_Variable_Number () {
	unassigned_variable_number--;
}


// returns the position of the variable var in the scope of the constraint, -1 if the variable does not belong to this scope
inline int Constraint::Get_Position (int var) {
	map<int,int>::iterator pos = position.find (var);

	if (pos == position.end())  return -1;
	else return pos->second;	
}


// returns true if the constraint has a specific propagator, false otherwise
// by default, there is no specific propagator
inline bool Constraint::Has_Specific_Propagator () {
	return false;
}


// returns true if the application of the propagator of the constraint deletes a value from the domain of at least one variable from the scope, false otherwise (is_modified[i] is true if the i th variable of the scope has been modified, false otherwise)
inline bool Constraint::Propagate (CSP * pb, Multiple_Residue_Support * ls, Deletion_Stack * ds, bool * & is_modified) {
	throw ("Error: there is no specific propagator");
}


// returns true if the application of the propagator of the constraint w.r.t. the variable var deletes a value from the domain of at least one variable from the scope, false otherwise (is_modified[i] is true if the i th variable of the scope has been modified, false otherwise)
// by default, there is no specific propagator
inline bool Constraint::Propagate (CSP * pb, Assignment & A, Multiple_Residue_Support * ls, Deletion_Stack * ds, bool * & is_modified) {
	throw ("Error: there is no specific propagator");
}
#endif
