/// \file 
/// \brief Definitions related to the Domain class

#ifndef DOMAIN_H
#define DOMAIN_H

using namespace std;

#include <map>
#include <light_vector.h>

class Domain {         /// This class implements enumerated domains
	protected :
  	unsigned int initial_size;  		            ///< initial size of the domain
		unsigned int size; 						              ///< current size of the domain
		Light_Vector<int> real_value;  		 					///< maps each value to its real value (useful for some predicate or global constraints)
		map<int,int> real_value_to_value;						///< maps each real value to its value (useful for some global constraints)
		Light_Vector<bool> current_state;  					///< current state of the domain (i.e. the remaining values after filtering)
		Light_Vector<int> remaining_value; 					///< the list of the remaining values (i.e. values which are not removed)
		Light_Vector<int> remaining_value_position; ///< the position of the remaining value in the vector remaining_value
		Light_Vector<int> growing_ordering; 				///< the growing ordering over real values (growing_ordering[i] provides the value at position i)
		Light_Vector<int> rev_growing_ordering; 		///< the reverse growing ordering over real values (rev_growing_ordering[i] provides the position of the value i)
		int ind_min;																///< the index of the current minimum
		int ind_max;																///< the index of the current maximum

	public :
		// constructors
		Domain ();							 ///< constructs an empty domain
		Domain (Domain & d);     ///< constructs a domain by copying the domain d
		
		// operator
		Domain & operator= (const Domain & d); ///< operator =
	
		// basic functions	
    bool Is_Empty ();                     ///< returns true if the domain is empty, false otherwise
    unsigned int Get_Initial_Size (); 	 	///< returns the initial size of the domain
		unsigned int Get_Size (); 	          ///< returns the current size of the domain
		bool Is_Present (int val);						///< returns true if the value val is present in the domain, false otherwise
		int Get_Remaining_Value (int i);			///< returns the number of i th remaining value
		int Get_Remaining_Real_Value (int i);	///< returns the real value corresponding to the i th remaining value
		void Add_Value (int real_val);				///< adds a new value related to val in the domain
		int Get_Real_Value (int num);					///< returns the real value corresponding to the value num
		int Get_Value (int real_val);					///< returns the value corresponding to the real value real_val if it exists, -1 otherwise
		void Filter_Value (int val);   				///< removes by filtering the value val
		void Unfilter_Value (int val); 				///< restores the value val deleted by filtering
		int Get_Min ();												///< returns the lowest value currently in the domain
		int Get_Max ();												///< returns the largest value currently in the domain
		int Next_Value (int i);								///< returns the available value following the i th value in the growing ordering if any, -1 otherwise
		int Previous_Value (int i);						///< returns the available value preceding the i th value in the growing ordering if any, -1 otherwise
};


//-----------------------------
// inline function definitions
//-----------------------------

// returns true if the domain is empty, false otherwise
inline bool Domain::Is_Empty () {
	return size == 0;
}


// returns the initial size of the domain
inline unsigned  int Domain::Get_Initial_Size () {
	return initial_size;
}


// returns the current size of the domain
inline unsigned int Domain::Get_Size () {
	return size;
}


// returns true if the value val is present in the domain, false otherwise
inline bool Domain::Is_Present (int val) {
	return current_state[val];	
}


// returns the value corresponding to the real value real_val if it exists, -1 otherwise
inline int Domain::Get_Value (int real_val) {
	map<int,int>::iterator iter = real_value_to_value.find (real_val);
	if (iter == real_value_to_value.end())	return -1;
	else return iter->second;
}


// returns the number of i th remaining value
inline int Domain::Get_Remaining_Value (int i) {
	return remaining_value[i];
}


// returns the real value corresponding to the i th remaining value
inline int Domain::Get_Remaining_Real_Value (int i) {
	return real_value[remaining_value[i]];
}


// returns the real value corresponding to the value num
inline int Domain::Get_Real_Value (int num) {
	return real_value[num];
}


// removes by filtering the value val
inline void Domain::Filter_Value (int val) {
 	current_state [val] = false;
 	size--;	

 	int pos = remaining_value_position[val];
 	int tmp = remaining_value[pos];
 	remaining_value[pos] = remaining_value[size];
 	remaining_value[size] = tmp;
 	
 	remaining_value_position[remaining_value[pos]] = pos;
 	remaining_value_position[val] = size; 	
}


// restores the value val deleted by filtering
inline void Domain::Unfilter_Value (int val) {
	current_state [val] = true;
	
	if (real_value[val] < real_value[growing_ordering[ind_min]])		ind_min = rev_growing_ordering[val];
		
	if (real_value[val] > real_value[growing_ordering[ind_max]])		ind_max = rev_growing_ordering[val];
	
	int pos = remaining_value_position[val];
 	int tmp = remaining_value[pos];
 	remaining_value[pos] = remaining_value[size];
 	remaining_value[size] = tmp;
 	
 	remaining_value_position[remaining_value[pos]] = pos;
 	remaining_value_position[val] = size;	
 	
 	size++;
}


// returns the lowest value currently in the domain
inline int Domain::Get_Min () {
	while (! current_state[growing_ordering[ind_min]])		ind_min++;
  
	return growing_ordering[ind_min];
}


// returns the largest value currently in the domain
inline int Domain::Get_Max () {
	while (! current_state[growing_ordering[ind_max]])	ind_max--;
  
	return growing_ordering[ind_max];
}


// returns the available value following the i th value in the growing ordering if any, -1 otherwise
inline int Domain::Next_Value (int i) {
	i = rev_growing_ordering[i] + 1;
	while (((unsigned) i < initial_size) && (! current_state[growing_ordering[i]]))		i++;
	
  if ((unsigned) i < initial_size)		return growing_ordering[i];
	else return -1;
}


// returns the available value preceding the i th value in the growing ordering if any, -1 otherwise
inline int Domain::Previous_Value (int i) {
	i = rev_growing_ordering[i] - 1;
	while ((i >= 0) && (! current_state[growing_ordering[i]]))		i--;
	
  if (i >= 0)		return growing_ordering[i];
	else return -1;
}
#endif
