#include "load_XCSP3Callbacks.h"
#include "predicate_constraint.h"
#include "binary_extension_constraint.h"
#include "positive_nary_extension_constraint.h"
#include "negative_nary_extension_constraint.h"

#include "less_than_weighted_sum_arithmetic_global_constraint.h"
#include "less_equal_weighted_sum_arithmetic_global_constraint.h"
#include "greater_than_weighted_sum_arithmetic_global_constraint.h"
#include "greater_equal_weighted_sum_arithmetic_global_constraint.h"
#include "equal_weighted_sum_arithmetic_global_constraint.h"
#include "not_equal_weighted_sum_arithmetic_global_constraint.h"
#include "in_weighted_sum_arithmetic_global_constraint.h"

#include "less_than_sum_arithmetic_global_constraint.h"
#include "less_equal_sum_arithmetic_global_constraint.h"
#include "greater_than_sum_arithmetic_global_constraint.h"
#include "greater_equal_sum_arithmetic_global_constraint.h"
#include "equal_sum_arithmetic_global_constraint.h"
#include "not_equal_sum_arithmetic_global_constraint.h"
#include "in_sum_arithmetic_global_constraint.h"

#include "constant_element_global_constraint.h"
#include "constant_index_element_global_constraint.h"
#include "element_global_constraint.h"
#include "index_element_global_constraint.h"

#include "all_different_global_constraint.h"


//----------------------------
// constructor and destructor
//----------------------------


// constructor
Load_XCSP3Callbacks::Load_XCSP3Callbacks(CSP * pb_ref) : XCSP3CoreCallbacks() {
	pb = pb_ref;
}


//-----------------
// basic functions
//-----------------


void Load_XCSP3Callbacks::buildVariableInteger(string id, int minValue, int maxValue) {
	Domain * D = new Domain ();

	for (int i = minValue; i <= maxValue; i++)  D->Add_Value (i);
		
	pb->Add_Variable (new Variable(pb->Get_N()+1,id,D));	
}


void Load_XCSP3Callbacks::buildVariableInteger(string id, vector<int> &values) {
	Domain * D = new Domain ();

	for(unsigned int i = 0; i < values.size() ; i++)  D->Add_Value (values[i]);
		
	pb->Add_Variable (new Variable(pb->Get_N()+1,id,D));
}


void Load_XCSP3Callbacks::buildConstraintExtension(string id, vector<XVariable *> list, vector<vector<int> > &tuples, bool support, bool hasStar) {
	// we first consider the scope of the constraint
	vector<Variable *> scope;
	
	for (vector<XVariable*>::iterator iter = list.begin(); iter != list.end(); iter++)
		scope.push_back (pb->Get_Variable((*iter)->id));
	
	// we create the constraint
	Extension_Constraint * c;
	unsigned int arity = list.size();
	
	if (list.size() == 2) 	c = new Binary_Extension_Constraint (scope,! support);
	else
    if (support)	c = new Positive_Nary_Extension_Constraint(scope);
    else c = new Negative_Nary_Extension_Constraint(scope);
	
	int t [arity];

	last_tuples.clear();

	for (vector<vector<int>>::iterator iter = tuples.begin(); iter != tuples.end(); iter++) {
		for (unsigned int i = 0; i < arity; i++) {
			unsigned int j = 0; 
			while ((j < scope[i]->Get_Domain()->Get_Initial_Size()) && (scope[i]->Get_Domain()->Get_Real_Value(j) != (*iter)[i]))   j++;
				
			t[i] = j;
		}

		if (support) c->Allow_Tuple (t);
		else c->Forbid_Tuple (t);
	
		// we save the tuples for the case where a similar constraint is defined after the current constraint
		last_tuples.push_back (*iter);
	}

	pb->Add_Constraint (c);
}


// removes values forbidden by a unary extension constraint 
void Filter_Domain (Variable * x, vector<int> &tuples, bool support){
	for (unsigned int i = 0; i < x->Get_Domain()->Get_Initial_Size(); i++)
		if (x->Get_Domain()->Is_Present(i)) {
			vector<int>::iterator iter = tuples.begin(); 
			while ((iter != tuples.end()) && (x->Get_Domain()->Get_Real_Value(i) != *iter))  iter++;
				
			if (iter == tuples.end()) {
				// the value i does not belong to the vector tuples
				if (support)	x->Get_Domain()->Filter_Value(i);
			}
			else if (! support)	x->Get_Domain()->Filter_Value(i);				// the value i belongs to the vector tuples
		}
}


void Load_XCSP3Callbacks::buildConstraintExtension(string id, XVariable *variable, vector<int> &tuples, bool support, bool hasStar) {
	// we first consider the scope of the constraint
	Variable * x;
	
	x = pb->Get_Variable(variable->id);

	
	// we remove forbidden values
	Filter_Domain (x, tuples,support);
	
	
	// we save the tuples for the case where a similar constraint is defined after the current constraint
	last_tuples.clear();
	last_tuples.push_back(tuples);
}


void Load_XCSP3Callbacks::buildConstraintExtensionAs(string id, vector<XVariable *> list, bool support, bool hasStar)  {
	// we first consider the scope of the constraint
	vector<Variable *> scope;
	
	for (vector<XVariable*>::iterator iter = list.begin(); iter != list.end(); iter++)
		scope.push_back (pb->Get_Variable((*iter)->id));
	
	// we create the constraint
	Extension_Constraint * c;
	unsigned int arity = list.size();
	
  if (list.size() == 1) 	Filter_Domain (scope[0], last_tuples[0],support);
  else {
      if (list.size() == 2)  c = new Binary_Extension_Constraint (scope,! support);
      else if (support)  c = new Positive_Nary_Extension_Constraint(scope);
           else c = new Negative_Nary_Extension_Constraint(scope);
      
      int t [arity];
      for (vector<vector<int>>::iterator iter = last_tuples.begin(); iter != last_tuples.end(); iter++) {
        for (unsigned int i = 0; i < arity; i++) {
          unsigned int j = 0; 
          while ((j < scope[i]->Get_Domain()->Get_Initial_Size()) && (scope[i]->Get_Domain()->Get_Real_Value(j) != (*iter)[i]))   j++;
            
          t[i] = j;
        }

        if (support)  c->Allow_Tuple (t);
        else c->Forbid_Tuple (t);
      }

      pb->Add_Constraint (c);
    }
}


// returns true if the string term corresponds to an operator (if so op will contain this operator), false otherwise
bool is_operator (string term, operator_type & op) {
  if (term.compare ("neg") == 0)  op = NEG;
  else if (term.compare ("abs") == 0)  op = ABS;
    else if (term.compare ("add") == 0)  op = ADD;
      else if (term.compare ("sub") == 0)  op = SUB;
        else if (term.compare ("mul") == 0)  op = MUL;
          else if (term.compare ("div") == 0)  op = DIV;
            else if (term.compare ("mod") == 0)  op = MOD;
              else if (term.compare ("pow") == 0)  op = POW;
                else if (term.compare ("min") == 0)  op = MIN;
                  else if (term.compare ("max") == 0)  op = MAX;
                    else if (term.compare ("eq") == 0)  op = EQ;
                      else if (term.compare ("ne") == 0)  op = NE;
                        else if (term.compare ("ge") == 0)  op = GE;
                         else if (term.compare ("gt") == 0)  op = GT;
                           else if (term.compare ("le") == 0)  op = LE;
                             else if (term.compare ("lt") == 0)  op = LT;
                               else if (term.compare ("not") == 0)  op = NOT;
                                 else if (term.compare ("and") == 0)  op = AND;
                                   else if (term.compare ("or") == 0)  op = OR;
                                     else if (term.compare ("xor") == 0)  op = XOR;
                                       else if (term.compare ("iff") == 0)  op = IFF;
                                         else if (term.compare ("if") == 0)  op = IF;
																					 else if (term.compare ("dist") == 0)  op = DIST;
																						 else if (term.compare ("sqr") == 0)  op = SQR;
                                               else if (term.compare ("imp") == 0)  op = IMP;
                                                 else if (term.compare ("set") == 0)  op = SET;
                                                   else if (term.compare ("in") == 0)  op = IN;
                                                     else return false;
  return true;
}

// returns the term located from the position pos
string extract_term (string expression, unsigned int & pos) {
	string term;
	while ((pos < expression.length()) && (expression[pos] != '(') && (expression[pos] != ',') && (expression[pos] != ')')) {
		term.push_back (expression[pos]);
		pos++;
	}
	
	return term;
}


// reads a functional predicate in the file whose descriptor is file and returns the corresponding expression
Expression string_to_expression (string expression, unsigned int & pos, CSP * pb, vector<Variable *> & scope) {
	string term;

	term = extract_term (expression,pos);

	if ((isdigit(term[0])) || (term[0] == '-'))  return Expression (VAL,stoi(term));   // the current expression is a value
	else {
			// the current expression corresponds to a variable or an operator
			
			operator_type op;
			if (is_operator (term,op)) {
				// the expression corresponds to an operator
				
				vector<Expression> subexpr;
				pos++;
				
				while (expression[pos] != ')') {
					subexpr.push_back (string_to_expression(expression,pos,pb,scope));
					
					if (expression[pos] == ',')  pos++;
				}
				pos++;
				
				if (subexpr.size() > 2) {
					// if the operator has more than two operands, we replace some operators by their n-ary version
					switch (op) {
						case ADD: op = SUM; break;
						case MUL: op = PROD; break;
						case EQ : op = EQU; break;
						default: ;		// for avoiding warning
					}
				}
				
				return Expression (op,subexpr);
			}
			else {
					// the expression is the name of a variable
					unsigned int i = 0;
					while ((i < scope.size()) && (scope[i]->Get_Name().compare (term) != 0))  i++;
					
					if (i == scope.size())	scope.push_back (pb->Get_Variable(term));
						
					return Expression (VAR,i);
				}
		}
}


Expression  string_to_expression (string expression, CSP * pb, vector<Variable *> & scope) {
	unsigned int pos = 0;
	return string_to_expression (expression, pos, pb, scope);
}


void Load_XCSP3Callbacks::buildConstraintIntension(string id, string expr) {
    vector<Variable *> scope;
    Expression relation = string_to_expression (expr,pb,scope);
        
    if (scope.size() == 1) {
			for (int i = 0; (unsigned) i < scope[0]->Get_Domain()->Get_Initial_Size(); i++)
				if (scope[0]->Get_Domain()->Is_Present(i)) {
          int real_val =  scope[0]->Get_Domain()->Get_Real_Value(i);
          
          if (relation.Evaluate(&real_val) == 0)  scope[0]->Get_Domain()->Filter_Value(i);
        }
		}
    else pb->Add_Constraint (new Predicate_Constraint (scope, relation));
}


void Load_XCSP3Callbacks::buildConstraintPrimitive(string id, OrderType op, XVariable *x, int k, XVariable *y) {
  string expression;
  
  switch (op) {
		case XCSP3Core::LT: expression = "lt"; break;
		case XCSP3Core::LE: expression = "le"; break;
		case XCSP3Core::GT: expression = "gt"; break;
		case XCSP3Core::GE: expression = "ge"; break;
		case XCSP3Core::EQ: expression = "eq"; break;
		case XCSP3Core::NE: expression = "ne"; break;
		default: throw ("Error: operator not allowed");
  }
  
  expression += "(sub("+x->id+","+y->id+"),"+to_string(-k)+")";
  
	vector<Variable *> scope;
	Expression relation = string_to_expression (expression,pb,scope);
	
	pb->Add_Constraint (new Predicate_Constraint (scope, relation));
}


// creates an all-diff global constraint
void Load_XCSP3Callbacks::buildConstraintAlldifferent(string id, vector<XVariable *> &list) {
	// we first consider the scope of the constraint
	vector<Variable *> scope;
	
	for (vector<XVariable*>::iterator iter = list.begin(); iter != list.end(); iter++)
		scope.push_back (pb->Get_Variable((*iter)->id));
	
	// we create the constraint
	pb->Add_Constraint (new All_Different_Global_Constraint (scope));
}


void Load_XCSP3Callbacks::buildConstraintSum(string id, vector<XVariable *> &list, vector<int> &coeffs, XCondition &cond) {  
	// we first consider the scope of the constraint
	vector<Variable *> scope;
	
	for (vector<XVariable*>::iterator iter = list.begin(); iter != list.end(); iter++)
		scope.push_back (pb->Get_Variable((*iter)->id));
		
	if (cond.operandType == XCSP3Core::VARIABLE) {
		scope.push_back (pb->Get_Variable(cond.var));
		coeffs.push_back (-1);
		cond.val = 0;
	}
		
	// we create the constraint
	switch (cond.op) {
		case XCSP3Core::EQ: pb->Add_Constraint (new Equal_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.val)); break;
		case XCSP3Core::NE: pb->Add_Constraint (new Not_Equal_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.val)); break;
		case XCSP3Core::LT: pb->Add_Constraint (new Less_Than_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.val)); break;
		case XCSP3Core::LE:	pb->Add_Constraint (new Less_Equal_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.val)); break;
		case XCSP3Core::GT:	pb->Add_Constraint (new Greater_Than_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.val)); break;
		case XCSP3Core::GE: pb->Add_Constraint (new Greater_Equal_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.val)); break;
		case XCSP3Core::IN: pb->Add_Constraint (new In_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.min,cond.max)); break;
	}
}


void Load_XCSP3Callbacks::buildConstraintSum(string id, vector<XVariable *> &list, XCondition &cond) {
	// we first consider the scope of the constraint
	vector<Variable *> scope;
	
	for (vector<XVariable*>::iterator iter = list.begin(); iter != list.end(); iter++)
		scope.push_back (pb->Get_Variable((*iter)->id));

	if (cond.operandType == XCSP3Core::VARIABLE) {
		vector<int> coeffs;
		
		for (vector<XVariable*>::iterator iter = list.begin(); iter != list.end(); iter++)
			coeffs.push_back (1);

		scope.push_back (pb->Get_Variable(cond.var));
		coeffs.push_back (-1);
		cond.val = 0;
				
		// we create the constraint
		switch (cond.op) {
			case XCSP3Core::EQ: pb->Add_Constraint (new Equal_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.val)); break;
			case XCSP3Core::NE: pb->Add_Constraint (new Not_Equal_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.val)); break;
			case XCSP3Core::LT: pb->Add_Constraint (new Less_Than_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.val)); break;
			case XCSP3Core::LE:	pb->Add_Constraint (new Less_Equal_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.val)); break;
			case XCSP3Core::GT:	pb->Add_Constraint (new Greater_Than_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.val)); break;
			case XCSP3Core::GE: pb->Add_Constraint (new Greater_Equal_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.val)); break;
			case XCSP3Core::IN: pb->Add_Constraint (new In_Weighted_Sum_Arithmetic_Global_Constraint (scope,coeffs,cond.min,cond.max)); break;
		}
	}
	else // we create the constraint
			switch (cond.op) {
				case XCSP3Core::EQ: pb->Add_Constraint (new Equal_Sum_Arithmetic_Global_Constraint (scope,cond.val)); break;
				case XCSP3Core::NE: pb->Add_Constraint (new Not_Equal_Sum_Arithmetic_Global_Constraint (scope,cond.val)); break;
				case XCSP3Core::LT: pb->Add_Constraint (new Less_Than_Sum_Arithmetic_Global_Constraint (scope,cond.val)); break;
				case XCSP3Core::LE:	pb->Add_Constraint (new Less_Equal_Sum_Arithmetic_Global_Constraint (scope,cond.val)); break;
				case XCSP3Core::GT:	pb->Add_Constraint (new Greater_Than_Sum_Arithmetic_Global_Constraint (scope,cond.val)); break;
				case XCSP3Core::GE: pb->Add_Constraint (new Greater_Equal_Sum_Arithmetic_Global_Constraint (scope,cond.val)); break;
				case XCSP3Core::IN: pb->Add_Constraint (new In_Sum_Arithmetic_Global_Constraint (scope,cond.min,cond.max)); break;
			}
}


void Load_XCSP3Callbacks::buildConstraintElement(string id, vector<XVariable *> &list,  int value)  {
	// we first consider the scope of the constraint
	vector<Variable *> scope;
  
	for (vector<XVariable*>::iterator iter = list.begin(); iter != list.end(); iter++)
		scope.push_back (pb->Get_Variable((*iter)->id));

  // we create the constraint
  pb->Add_Constraint (new Constant_Element_Global_Constraint (scope,value));
}


void Load_XCSP3Callbacks::buildConstraintElement(string id, vector<XVariable *> &list, XVariable *value) {
  // we first consider the scope of the constraint
	vector<Variable *> scope;
  
	for (vector<XVariable*>::iterator iter = list.begin(); iter != list.end(); iter++)
		scope.push_back (pb->Get_Variable((*iter)->id));
  scope.push_back (pb->Get_Variable(value->id));

  // we create the constraint
  pb->Add_Constraint (new Element_Global_Constraint (scope));
}


void Load_XCSP3Callbacks::buildConstraintElement(string id, vector<XVariable *> &list, int startIndex, XVariable *index, RankType rank, int value)  {
	// we first consider the scope of the constraint
	vector<Variable *> scope;
	
  scope.push_back (pb->Get_Variable(index->id));
	for (vector<XVariable*>::iterator iter = list.begin(); iter != list.end(); iter++)
		scope.push_back (pb->Get_Variable((*iter)->id));

  // we create the constraint
  pb->Add_Constraint (new Constant_Index_Element_Global_Constraint (scope,value));
}


void Load_XCSP3Callbacks::buildConstraintElement(string id, vector<XVariable *> &list, int startIndex, XVariable *index, RankType rank, XVariable *value)  {
  // we first consider the scope of the constraint
	vector<Variable *> scope;
  
  scope.push_back (pb->Get_Variable(index->id));
	for (vector<XVariable*>::iterator iter = list.begin(); iter != list.end(); iter++)
		scope.push_back (pb->Get_Variable((*iter)->id));
  scope.push_back (pb->Get_Variable(value->id));

  // we create the constraint
  pb->Add_Constraint (new Index_Element_Global_Constraint (scope));
}
