#include <iostream>
#include <string>

// AC header files
#include "constraint_ac.h"
#include "variable_nogood_base_ac.h"

// deletion_stack header files
#include "basic_deletion_stack.h"
#include "support_deletion_stack.h"


// solver header files
#include "solver.h"

// general header files
#include "general_counters.h"
#include "auxiliary_functions_main.h"

#include <fstream>
#include <XCSP3CoreParser.h>
#include "load_XCSP3Callbacks.h"

// tree-decomposition header files

#include "max_ratio_root_heuristic.h"
#include "max_weight_sum_root_heuristic.h"



// prints the information related to the solving
void Print_Solving_Information (CSP * pb,int solver_result,  Assignment solution) {
	switch (solver_result) {
		case -1 : // the solving is stopped due to the timeout 
							cout << "s UNKNOWN" << endl;	break;
		case  0 : // there is no solution
							cout << "s UNSATISFIABLE" << endl;	break;
		case  1 : // there is a solution
							cout << "s SATISFIABLE" << endl;
              cout << "v <instantiation type=\"solution\">" << endl;
              cout << "v <list>";
              for (unsigned int i = 0; i < pb->Get_N(); i++)
                cout << " " << pb->Get_Variable(i)->Get_Name();
              cout << " </list>" << endl;
              cout << "v <values>";
              for (unsigned int i = 0; i < pb->Get_N(); i++)
                cout << " " << pb->Get_Variable(i)->Get_Domain()->Get_Real_Value (solution.Get_Value (i));
              cout << " </values>" << endl;
              cout << "v </instantiation>" << endl;
							
							break;
	}
		
  cout << "c Solving # Nodes: " << Solving_Node_Counter.Get_Value() << endl;
}


// initialization of AC filtering
AC * AC_initialization (CSP * pb, int ac_version) {
	Constraint_Selection * cs = new Constraint_Selection;
  Multiple_Residue_Support * ls = new Multiple_Residue_Support (pb);

	if (ac_version == 0)  return new Constraint_AC (pb,ls, cs, new Filtering_Data_Structure<pair<int,int> >,false); // constraint oriented AC
  else return new Variable_Nogood_Base_AC (pb,ls,cs,new Filtering_Data_Structure<int>,false);     // variable oriented AC
}


// initialization of the solver
Solver * solver_initialization (CSP * pb, double time_limit, Tree_Decomposition * & TD) {	
	// initialization of the variable heuristic
	Variable_Heuristic * var_heur =  new Variable_Heuristic(pb->Get_N());
	
	// initialization of the value heuristic
	Value_Heuristic * val_heur = new Value_Heuristic();
	
	// initialization of elements specific to BTD
	
	int nb_clusters;

	Graphical_Structure * gs = pb->Get_Structure();
  Light_Graph * g = gs->Get_Primal_Light_Graph ();	
	TD = new Tree_Decomposition (gs,new First_Cluster_Heuristic(gs, 100000),new Next_Vertices_Heuristic (g, 100000));
		
	nb_clusters = TD->Nb_Clusters();

  Cluster_Heuristic * next_cluster_heur = new Cluster_Heuristic (nb_clusters);
  Root_Heuristic * root_cluster_heur = new Max_Ratio_Root_Heuristic;
  Root_Heuristic * restart_root_cluster_heur = new Max_Weight_Sum_Root_Heuristic;;

		
	// initialization of the solver
	Solver * solver;
	Deletion_Stack * ds;

	try {
		int size = 0;
    for (unsigned int i = 0; i < pb->Get_N(); i++)
      size += pb->Get_Variable(i)->Get_Domain()->Get_Size();
    
    ds = new Basic_Deletion_Stack(pb,size);
		
    solver = new Solver (pb,var_heur, val_heur,AC_initialization (pb, 1), ds, TD, root_cluster_heur, next_cluster_heur, new Restart_Policy (100,1.1), 
                         restart_root_cluster_heur, new Merge_Policy (50,1,100,TD->Get_Max_Label()+1), time_limit);
	}
	
	catch (...) {
		delete next_cluster_heur;
		
		throw;
	}
	delete next_cluster_heur;
	
	return solver;
}


int main (int argc, char * argv []) {
	try {		
		Solving_Timer.Start();

		if (argc != 3) {
      cerr << "Usage: " << argv[0] << " TIMELIMIT BENCHNAME" << endl;
      return 1;
    }
  
		CSP * pb = new CSP();
 	
		// read of the file
		try {
      ifstream file (argv[argc-1], ios::in);
  
      Load_XCSP3Callbacks cb (pb); 	// creates an interface between the parser and the current library

      XCSP3Core::XCSP3CoreParser parser(&cb);
      parser.parse(file);
      
      file.close();
		}

		catch (char const * error) {
			string s (error);
			if (s.compare("Unsupported") == 0)	cout << "s UNSUPPORTED" << endl;
			throw;
		}

    double time_limit = atof(argv[argc-2]);								// the limit of the solver runtime		
    
		AC * preprocessing = 0;
		Solver * solver = 0;
		Tree_Decomposition * TD = 0;
		
		try {
			cout << "c Version: 2017-08-10" << endl;

      // we replace some constraints in intention by constraints in extension
      Replace_Predicate_By_Global_Constraint (pb);					
			
			// we perform the preprocessing
			preprocessing = AC_initialization (pb,0);
			
      bool preprocessing_result = preprocessing->Achieve (pb);
			
			// initialization of the solver
			int result = 0;
			
			Assignment solution (pb->Get_N());
						
			if (preprocessing_result) {
        solver = solver_initialization (pb,time_limit,TD);

        result = solver->Solve ();
                  
        if (result == 1)  solution = solver->Get_Solution();
			}

			// print the information related to the solving
			Print_Solving_Information (pb,result,solution);
		
      Solving_Timer.Stop();
			cout << "c Global Time: " << Solving_Timer.Get_Duration() << " s" << endl;
		}
		
		catch (...) {
			delete preprocessing;
			delete solver;
			delete TD;
			delete pb;
			throw;
		}
		
		delete preprocessing;
		delete solver;
		delete TD;
		delete pb;
	}
 	
 	catch (...){
		return 1;
  }

  return 0;
}
