/// \file 
/// \brief Definitions related to the Solver class

#ifndef SOLVER_H
#define SOLVER_H

using namespace std;

#include "root_heuristic.h"
#include "cluster_heuristic.h"
#include "variable_heuristic.h"
#include "value_heuristic.h"
#include "csp_tree_decomposition.h"
#include "restart_policy.h"
#include "ac.h"
#include "csp_child_node.h"
#include "merge_policy.h"

enum Solver_State { ///< type for representing the state of solver
  NOT_RUN,        ///< the solver has not been run yet
  FINISHED,       ///< the solver has been run and has finished
  STOPPED         ///< the solver has been run and has been stopped
};     


class Solver { 		/// this class implements the algorithm BTD-MAC-Merge for solving CSP instances
	protected:
    Solver_State state;													  ///< the current state of the solver
		Variable_Heuristic * variable_heuristic;			///< the variable heuristic used in the solver for ordering the variables and so for choosing the next variable to be assigned
		Value_Heuristic * value_heuristic;						///< the value heuristic used in the solver for ordering the values of each variable
		Deletion_Stack * ds;													///< the stack of value deletion
		double time_limit;														///< the time limit 
		Assignment A;																	///< the current assignment of the solver
		CSP * pb;																			///< the instance to solve
  
    bool has_found_solution; 		///< true if the solver has found a solution, false otherwise
  
		CSP_Tree_Decomposition * td;											///< the tree-decomposition used for the solving
		Root_Heuristic * root_heuristic;									///< the root heuristic used in the solver for choosing the first cluster
		Cluster_Heuristic ** cluster_heuristic;						///< the array of cluster heuristic used in the solver for choosing the next cluster to study (we consider an heuristic instance per cluster)
		Variable_Heuristic ** cluster_variable_heuristic;	///< the array of variableheuristic used in the solver for choosing the next variable to assign (we consider an heuristic instance per cluster)
    
    AC * ac;																		///< the AC filtering
		Restart_Policy * restart_policy;						///< the global restart policy
		Root_Heuristic * restart_root_heuristic;		///< the heuristic used at each restart for choosing the new root cluster
  
		Merge_Policy * merge_policy; 										///< the policy used for merging clusters
		Variable_Heuristic * simulation;								///< the variable heuristic used to check whether BTD-MAC-Merge prefers a variable from the current cluster rather one of its descent			
		CSP_Tree_Decomposition * original_td_merge; 		///< the orignal considered tree-decomposition (i.e. before any merging)
		

	public:
		// constructor and destructor
		Solver (CSP * inst, Variable_Heuristic * var_heur, Value_Heuristic * val_heur, AC * cc, Deletion_Stack * stack, Tree_Decomposition * TD, 
            Root_Heuristic * root_heur, Cluster_Heuristic * cluster_heur, Restart_Policy *rp, Root_Heuristic * rs_root_heur, Merge_Policy * local_merge_p, 
            double limit);	///< constructs a solver in order to solve the instance inst by using the variable heuristic var_heur, the value heuristic val_heur, the consistency check,the deletion stack stack, the tree-decomposition TD, the root cluster heuristic root_heur, the cluster heuristic cluster_heur, the global restart policy rp, the local restart policy local_rp, the cluster root heuristic rs_root_heur for choosing the root cluster at each restart and the policy local_merge_p for merging clusters. The runtime is limited to limit. Child nodes rely on lexicographical trees.
		~Solver ();			///< destructor
		

		// basic functions
		int Solve ();								      	///< returns 0 if the CSP is inconsistent, 1 if it is consistent, -1 if the solving is interrupted (the runtime exceeds the time limit)
    Solver_State Get_Solver_State ();		///< returns the current solver's state

		Assignment & Get_Solution ();	      ///< returns the solution if the CSP is consistent or an empty assignment if the CSP is inconsistent or has not been solved yet		
};


//-----------------------------
// inline function definitions
//----------------------------


// returns the current solver's state
inline Solver_State Solver::Get_Solver_State () {
	return state;
}


#endif

