#include "first_cluster_heuristic.h"
#include "light_graph.h"
#include "connected_component.h"


//----------------------------
// constructor and destructor
//----------------------------

// constructs a heuristic which chooses a clique (of the primal graph of the graphical structure gs)
First_Cluster_Heuristic::First_Cluster_Heuristic (Graphical_Structure * gs, int sep) {
	g = gs;
	k = sep;
}


//-----------------
// basic functions
//-----------------


// returns the first cluster for a tree-decomposition according to a heuristic method. This cluster must be connected and its vertices must be chosen among the vertices for which the flag is_available is set to true. For each selected vertex, is_available is set to false
Cluster_Node * First_Cluster_Heuristic::Choose (bool is_available[]) {
	Light_Graph * pg = g->Get_Primal_Light_Graph();
	
	unsigned int n = pg->N();
	int deg, min_degree;
	vertex min_v = 0;
				
	min_degree = INT_MAX;
	for (vertex v = 0; v < n; v++)
		if (is_available[v]) {
			deg = g->Degree (v);
			if (deg < min_degree) {
				min_degree = deg;
				min_v = v;
			}
		}
	
	Light_List<vertex>::iterator iter_v;
	
	pair<int, vertex *> sep_k_1; // level k+1
	sep_k_1.first = 0;
	sep_k_1.second = new vertex [g->N()];
	
	pair<int, vertex *> sep_k; // level k
	sep_k.first = 1;
	sep_k.second = new vertex [g->N()];
	sep_k.second[0] = min_v;
	
	// the traversal
	Cluster_Node * cluster_t = new Cluster_Node (g);
	cluster_t->Add_Vertex (min_v);
	
	int * belong_to_traversed = new int [g->N()];
	
	for (unsigned int i=0; i<g->N(); i++)  belong_to_traversed[i] = 0;
		
	belong_to_traversed[min_v]= 1; 
	
	// the cluster to return
	Cluster_Node * cluster = new Cluster_Node (g);
	pair<int, vertex *> cluster_taken;
	cluster_taken.first = 0;
	cluster_taken.second = new vertex [g->N()];

	bool end = false;
	
	// search for the smallest separator from min_v and <= k
	while (!end && !( k >= 1 && cluster_taken.first == 1)) {
		sep_k_1.first = 0; // level k+1
		
		for (int i = 0; i < sep_k.first; i++) //level k
			// stepping one level
			for (iter_v = pg->Begin_Neighbor (sep_k.second[i]); iter_v != pg->End_Neighbor (sep_k.second[i]); iter_v++)
				if(belong_to_traversed[*iter_v] == 0)	{	
					belong_to_traversed[*iter_v] = 1;
				
					cluster_t->Add_Vertex (*iter_v);
				
					sep_k_1.second[sep_k_1.first] = *iter_v;
					sep_k_1.first++;				
				}
		
		if (sep_k_1.first == 0)  end = true;
		else	{
			Connected_Component CC (g);
			
			Light_Stack<tuple<pair<int,vertex *>,Cluster_Node *,pair<int,vertex *>>> S (n);
			
			CC.Get_Connected_Components_From_Set (cluster_t,S);
					
			pair<int,vertex *> V_i;
			pair<int,vertex *> X_i;
			
			while (! S.is_empty()) {
				// we consider the neighborhood V_i 
				X_i = get<0>(S.top());
				V_i = get<2>(S.top());
				S.pop();
				
				// good separator found
				if (V_i.first <= k && (V_i.first < cluster_taken.first || cluster_taken.first == 0))	{	
					cluster_taken.first = 0;
					for(int i=0; i< V_i.first; i++)	{	
						cluster_taken.second[cluster_taken.first] = V_i.second[i];
						cluster_taken.first ++;
					}	
				}
				
				delete [] V_i.second;
				delete [] X_i.second;
			}
			
			sep_k.first = sep_k_1.first;
			for(int i=0; i< sep_k_1.first; i++)  sep_k.second[i] = sep_k_1.second[i];
		}
	}
	
	delete [] sep_k.second;
	delete [] sep_k_1.second;
	delete [] belong_to_traversed;
	delete cluster_t;
	
	if (cluster_taken.first == 0)	{
		cluster->Add_Vertex (min_v);
		is_available[min_v] = false;
	}
	else
		for(int i=0; i< cluster_taken.first; i++)	{
			cluster->Add_Vertex (cluster_taken.second[i]);
			is_available[cluster_taken.second[i]] = false;
		}
	
	delete [] cluster_taken.second;		
	
	return cluster;
}
