/*=============================================================================
 * parser for CSP instances represented in XCSP3 Format
 *
 * Copyright (c) 2015 xcsp.org (contact <at> xcsp.org)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *=============================================================================
 */
#ifndef COSOCO_XCSP3CORECALLBACKS_H
#define COSOCO_XCSP3CORECALLBACKS_H

#include "XCSP3Constants.h"
#include "XCSP3Variable.h"
#include "XCSP3Constraint.h"
#include <vector>
#include <string>


namespace XCSP3Core {

    using namespace std;

    class XCSP3CoreCallbacks {
    protected :
        vector<string> classesToDiscard;
    public :

        bool recognizeSpecialIntensionCases; // x<y.. x=k,
        bool recognizeSpecialCountCases;     // atleast, atmost, exactly, among, exctalyVariable
        bool recognizeNValuesCases;          // AllEqual, NotAllEqual, AllDiff
        bool normalizeSum;                   // Normalize sum : merge same variables...remove var with coef = 0,

        XCSP3CoreCallbacks() {
            recognizeSpecialIntensionCases = true;
            recognizeSpecialCountCases = true;
            recognizeNValuesCases = true;
            normalizeSum = true;
        }


        void addClassToDiscard(string cl) {
            classesToDiscard.push_back(cl);
        }


        bool discardedClasses(string classes) {
            if(classes == "")
                return false;

            for(string c : classesToDiscard)
                if(classes.find(c) != std::string::npos)
                    return true;
            return false;
        }


        // All these callbacks are called when the tag starts and when it ends.
        virtual void beginInstance(InstanceType type) {}


        virtual void endInstance() {}


        virtual void beginVariables() {}


        virtual void endVariables() {}


        virtual void beginVariableArray(string id) {} //beginArray
        virtual void endVariableArray() {}


        virtual void beginConstraints() {}


        virtual void endConstraints() {}


        virtual void beginGroup(string id) {}


        virtual void endGroup() {}


        virtual void beginBlock(string classes) {}


        virtual void endBlock() {}


        virtual void beginSlide(string id, bool circular) {}


        virtual void endSlide() {}


        virtual void beginObjectives() {}


        virtual void endObjectives() {}


        //--------------------------------------------------------------------------------------
        // Build Variable. Must be implemented.
        //--------------------------------------------------------------------------------------
        virtual void buildVariableInteger(string id, int minValue, int maxValue) = 0;

        virtual void buildVariableInteger(string id, vector<int> &values) = 0;



        //--------------------------------------------------------------------------------------
        // Basic constraints
        //--------------------------------------------------------------------------------------

        virtual void buildConstraintExtension(string id, vector<XVariable *> list, vector <vector<int>> &tuples, bool support, bool hasStar) {
            throw ("Unsupported");
        }


        virtual void buildConstraintExtension(string id, XVariable *variable, vector<int> &tuples, bool support, bool hasStar) {
            throw ("Unsupported");
        }


        virtual void buildConstraintExtensionAs(string id, vector<XVariable *> list, bool support, bool hasStar) {
            throw ("Unsupported");
        }


        virtual void buildConstraintIntension(string id, string expr) {
            throw ("Unsupported");
        }


        virtual void buildConstraintPrimitive(string id, OrderType op, XVariable *x, int k, XVariable *y) {
            throw ("Unsupported");
        }

        //--------------------------------------------------------------------------------------
        // Language constraints
        //--------------------------------------------------------------------------------------

        virtual void buildConstraintRegular(string id, vector<XVariable *> &list, string st, vector<string> &final, vector<XTransition> &transitions) {
            throw ("Unsupported");
        }


        virtual void buildConstraintMDD(string id, vector<XVariable *> &list, vector<XTransition> &transitions) {
            throw ("Unsupported");
        }


        //--------------------------------------------------------------------------------------
        // Comparison constraints
        //--------------------------------------------------------------------------------------

        virtual void buildConstraintAlldifferent(string id, vector<XVariable *> &list) {
            throw ("Unsupported");
        }


        virtual void buildConstraintAlldifferentExcept(string id, vector<XVariable *> &list, vector<int> &except) {
            throw ("Unsupported");
        }


        virtual void buildConstraintAlldifferentList(string id, vector<vector<XVariable *>> &lists) {
            throw ("Unsupported");
        }


        virtual void buildConstraintAlldifferentMatrix(string id, vector<vector<XVariable *>> &matrix) {
            throw ("Unsupported");
        }


        virtual void buildConstraintAllEqual(string id, vector<XVariable *> &list) {
            throw ("Unsupported");
        }


        virtual void buildConstraintNotAllEqual(string id, vector<XVariable *> &list) {
            throw ("Unsupported");
        }


        virtual void buildConstraintOrdered(string id, vector<XVariable *> &list, OrderType order) {
            throw ("Unsupported");
        }


        virtual void buildConstraintLex(string id, vector<vector<XVariable *>> &lists, OrderType order) {
            throw ("Unsupported");
        }


        virtual void buildConstraintLexMatrix(string id, vector<vector<XVariable *>> &matrix, OrderType order) {
            throw ("Unsupported");
        }

        //--------------------------------------------------------------------------------------
        // Summing and counting constraints
        //--------------------------------------------------------------------------------------

        virtual void buildConstraintSum(string id, vector<XVariable *> &list, vector<int> &coeffs, XCondition &cond) {
            throw ("Unsupported");
        }


        virtual void buildConstraintSum(string id, vector<XVariable *> &list, XCondition &cond) {
            throw ("Unsupported");
        }


        virtual void buildConstraintSum(string id, vector<XVariable *> &list, vector<XVariable *> &coeffs, XCondition &cond) {
            throw ("Unsupported");
        }


        virtual void buildConstraintAtMost(string id, vector<XVariable *> &list, int value, int k) {
            throw ("Unsupported");
        }


        virtual void buildConstraintAtLeast(string id, vector<XVariable *> &list, int value, int k) {
            throw ("Unsupported");
        }


        virtual void buildConstraintExactlyK(string id, vector<XVariable *> &list, int value, int k) {
            throw ("Unsupported");
        }


        virtual void buildConstraintExactlyVariable(string id, vector<XVariable *> &list, int value, XVariable *x) {
            throw ("Unsupported");
        }


        virtual void buildConstraintAmong(string id, vector<XVariable *> &list, vector<int> &values, int k) {// TODO AMONG
            throw ("Unsupported");
        }


        virtual void buildConstraintCount(string id, vector<XVariable *> &list, vector<int> &values, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCount(string id, vector<XVariable *> &list, vector<XVariable *> &values, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void buildConstraintNValues(string id, vector<XVariable *> &list, vector<int> &except, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void buildConstraintNValues(string id, vector<XVariable *> &list, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCardinality(string id, vector<XVariable *> &list, vector<int> values, vector<int> &occurs, bool closed) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCardinality(string id, vector<XVariable *> &list, vector<int> values, vector<XVariable *> &occurs, bool closed) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCardinality(string id, vector<XVariable *> &list, vector<int> values, vector<XInterval> &occurs, bool closed) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCardinality(string id, vector<XVariable *> &list, vector<XVariable *> values, vector<int> &occurs, bool closed) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCardinality(string id, vector<XVariable *> &list, vector<XVariable *> values, vector<XVariable *> &occurs, bool closed) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCardinality(string id, vector<XVariable *> &list, vector<XVariable *> values, vector<XInterval> &occurs, bool closed) {
            throw ("Unsupported");
        }
        //--------------------------------------------------------------------------------------
        // Connection constraints
        //--------------------------------------------------------------------------------------

        virtual void buildConstraintMinimum(string id, vector<XVariable *> &list, XCondition &xc) {
            throw ("Unsupported");

        }


        virtual void buildConstraintMinimum(string id, vector<XVariable *> &list, XVariable *index, int startIndex, RankType rank, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void buildConstraintMaximum(string id, vector<XVariable *> &list, XCondition &xc) {
            throw ("Unsupported");

        }


        virtual void buildConstraintMaximum(string id, vector<XVariable *> &list, XVariable *index, int startIndex, RankType rank, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void buildConstraintElement(string id, vector<XVariable *> &list, int value) {
            throw ("Unsupported");
        }


        virtual void buildConstraintElement(string id, vector<XVariable *> &list, XVariable *value) {
            throw ("Unsupported");
        }


        virtual void buildConstraintElement(string id, vector<XVariable *> &list, int startIndex, XVariable *index, RankType rank, int value) {
            throw ("Unsupported");
        }


        virtual void buildConstraintElement(string id, vector<XVariable *> &list, int startIndex, XVariable *index, RankType rank, XVariable *value) {
            throw ("Unsupported");
        }


        virtual void buildConstraintElement(string id, vector<int> &list, int startIndex, XVariable *index, RankType rank, XVariable *value) {
            throw ("Unsupported");
        }


        virtual void buildConstraintChannel(string id, vector<XVariable *> &list, int startIndex) {
            throw ("Unsupported");
        }


        virtual void buildConstraintChannel(string id, vector<XVariable *> &list1, int startIndex1, vector<XVariable *> &list2, int startIndex2) {
            throw ("Unsupported");
        }


        virtual void buildConstraintChannel(string id, vector<XVariable *> &list, int startIndex, XVariable *value) {
            throw ("Unsupported");
        }

        //--------------------------------------------------------------------------------------
        // packing and schedulling constraints
        //--------------------------------------------------------------------------------------

        virtual void buildConstraintStretch(string id, vector<XVariable *> &list, vector<int> &values, vector<XInterval> &widths) {
            throw ("Unsupported");
        }


        virtual void buildConstraintStretch(string id, vector<XVariable *> &list, vector<int> &values, vector<XInterval> &widths, vector<vector<int>> &patterns) {
            throw ("Unsupported");
        }


        virtual void buildConstraintNoOverlap(string id, vector<XVariable *> &origins, vector<int> &lengths, bool zeroIgnored) {
            throw ("Unsupported");
        }


        virtual void buildConstraintNoOverlap(string id, vector<XVariable *> &origins, vector<XVariable *> &lengths, bool zeroIgnored) {
            throw ("Unsupported");
        }


        virtual void buildConstraintNoOverlap(string id, vector<vector<XVariable *>> &origins, vector<vector<int>> &lengths, bool zeroIgnored) {
            throw ("Unsupported");
        }


        virtual void buildConstraintNoOverlap(string id, vector<vector<XVariable *>> &origins, vector<vector<XVariable *>> &lengths, bool zeroIgnored) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCumulative(string id, vector<XVariable *> &origins, vector<int> &lengths, vector<int> &heights, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCumulative(string id, vector<XVariable *> &origins, vector<int> &lengths, vector<XVariable *> &varHeights, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCumulative(string id, vector<XVariable *> &origins, vector<XVariable *> &lengths, vector<int> &heights, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCumulative(string id, vector<XVariable *> &origins, vector<XVariable *> &lengths, vector<XVariable *> &heights, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCumulative(string id, vector<XVariable *> &origins, vector<int> &lengths, vector<int> &heights, vector<XVariable *> &ends, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void
        buildConstraintCumulative(string id, vector<XVariable *> &origins, vector<int> &lengths, vector<XVariable *> &varHeights, vector<XVariable *> &ends, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void
        buildConstraintCumulative(string id, vector<XVariable *> &origins, vector<XVariable *> &lengths, vector<int> &heights, vector<XVariable *> &ends, XCondition &xc) {
            throw ("Unsupported");
        }


        virtual void
        buildConstraintCumulative(string id, vector<XVariable *> &origins, vector<XVariable *> &lengths, vector<XVariable *> &heights, vector<XVariable *> &ends, XCondition &xc) {
            throw ("Unsupported");
        }

        //--------------------------------------------------------------------------------------
        // instantiation constraints
        //--------------------------------------------------------------------------------------

        virtual void buildConstraintInstantiation(string id, vector<XVariable *> &list, vector<int> &values) {
            throw ("Unsupported");
        }

        //--------------------------------------------------------------------------------------
        // Graph constraints
        //--------------------------------------------------------------------------------------

        virtual void buildConstraintCircuit(string id, vector<XVariable *> &list, int startIndex) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCircuit(string id, vector<XVariable *> &list, int startIndex, int size) {
            throw ("Unsupported");
        }


        virtual void buildConstraintCircuit(string id, vector<XVariable *> &list, int startIndex, XVariable *size) {
            throw ("Unsupported");
        }

        //--------------------------------------------------------------------------------------
        // Objectives
        //--------------------------------------------------------------------------------------

        virtual void buildObjectiveMinimizeExpression(string expr) {
            throw ("Unsupported");
        }


        virtual void buildObjectiveMaximizeExpression(string expr) {
            throw ("Unsupported");
        }


        virtual void buildObjectiveMinimizeVariable(XVariable *x) {
            throw ("Unsupported");
        }


        virtual void buildObjectiveMaximizeVariable(XVariable *x) {
            throw ("Unsupported");
        }


        virtual void buildObjectiveMinimize(ExpressionObjective type, vector<XVariable *> &list, vector<int> &coefs) {
            throw ("Unsupported");
        }


        virtual void buildObjectiveMaximize(ExpressionObjective type, vector<XVariable *> &list, vector<int> &coefs) {
            throw ("Unsupported");
        }


        virtual void buildObjectiveMinimize(ExpressionObjective type, vector<XVariable *> &list) {
            throw ("Unsupported");
        }


        virtual void buildObjectiveMaximize(ExpressionObjective type, vector<XVariable *> &list) {
            throw ("Unsupported");
        }

    };


};


#endif //COSOCO_XCSP3CORECALLBACKS_H





























