#include "tree_decomposition.h"
#include <iostream>
#include <queue>
#include <stack>
#include <light_stack.h>
#include "connected_component.h"

//-----------------------------
// constructors and destructor
//-----------------------------


// constructs an empty tree-decomposition related to the graphical structure gs
Tree_Decomposition::Tree_Decomposition (Graphical_Structure * gs) {
	g = gs;
	ct_number = 0;
  first_cluster = 0;
	next_vertices = 0;
}


// constructs a tree-decomposition related to the graphical structure gs by exploiting the heuristic fc for computing the first cluster and the heuristic nv for choosing the next vertices to consider
// this tree decomposition is done without triangulation
Tree_Decomposition::Tree_Decomposition (Graphical_Structure * gs, First_Cluster_Heuristic * fch, Next_Vertices_Heuristic * nvh) {
	g = gs;
	ct_number = 0;
	first_cluster = fch;
	next_vertices = nvh;
	
  Connected_Component CC (g);	
	int n = g->N();
	bool * is_untreated = new bool [n];
	int remaining_vertices = n;
	
	for (int i = 0; i < n; i++)		is_untreated[i] = true;
	
	int current_w = -1;
	
	do {
		// we take into account a new cluster tree
		if ((unsigned) ct_number+1 >= cluster_tree.get_capacity ())		cluster_tree.expand(50);
		cluster_tree[ct_number] = new Cluster_Tree(g);

		// we compute the first cluster
		Cluster_Node * cluster;
		
		cluster = first_cluster->Choose (is_untreated);

		if (cluster->Size()-1 > current_w)	current_w = cluster->Size() - 1; 
			
		remaining_vertices -= cluster->Size();
		
		cluster_tree[ct_number]->Add_Cluster (cluster);

		// we now process the remaining part of the current connected component (if any)
		if (remaining_vertices > 0)	{
			Light_Stack<tuple<pair<int,vertex *>,Cluster_Node *,pair<int,vertex *>>> S (n);
			
			CC.Get_Connected_Components_From_Set (cluster,S);
					
			pair<int,vertex *> X_i;
			pair<int,vertex *> V_i;

			Light_List<vertex>::iterator iter;
			Light_List<Cluster_Node *>::iterator iter_cluster;
			Cluster_Node * parent;
			
			while (! S.is_empty()) {
				// we consider a new set X_i with its parent cluster and its neighborhood V_i 
				X_i = get<0>(S.top());
				parent = get<1>(S.top());
				V_i = get<2>(S.top());
				S.pop();
				
				// we check if V_i corresponds to an existing cluster	
				if (V_i.first == parent->Size())  cluster = parent;		// V_i and parent represent the same set of vertices. So we add the vertices of E_i \ V_i in parent
				else {
						// we add E_i as a new cluster
						cluster = new Cluster_Node (g);
						for (int i = 0; i < V_i.first; i++) 	cluster->Add_Vertex (V_i.second[i]);

						cluster_tree[ct_number]->Add_Cluster (cluster);
						cluster_tree[ct_number]->Add_Child_Node (parent,cluster);
					}
				
				// we construct the new cluster
				next_vertices->Choose (X_i,cluster,V_i,current_w,is_untreated);
				
				remaining_vertices -= cluster->Size() - V_i.first;
					
				if (cluster->Size() -1 > current_w)  current_w = cluster->Size() - 1; 

				// we compute the connected components of g [X_i \ E_i] and their neighborhood
				CC.Get_Connected_Components_With_Neighborhood (X_i,cluster,S);
			
				delete [] X_i.second;
				delete [] V_i.second;
			}
		}
		ct_number++;
	}
	while (remaining_vertices > 0);

	delete [] is_untreated;	
}


// destructor
Tree_Decomposition::~Tree_Decomposition () {
	for (int i = 0; i < ct_number; i++) 	delete cluster_tree[i];
    
  delete first_cluster;
	delete next_vertices;
}


//-----------------
// basic functions 
//-----------------


// returns the number of clusters of the cluster tree
int Tree_Decomposition::Nb_Clusters () {
	int sum = 0;
	
	for (int i = 0; i < ct_number; i++)		sum += cluster_tree[i]->Nb_Clusters ();
	
	return sum;
}


// returns the largest label of clusters in a cluster tree
int Tree_Decomposition::Get_Max_Label () {
	int max = 0;
	
	for (int i = 0; i < ct_number; i++)
		if (max < cluster_tree[i]->Get_Max_Label ())	max = cluster_tree[i]->Get_Max_Label ();
			
	return max;
}


// returns the largest number of vertices in a cluster tree
int Tree_Decomposition::Get_Max_Vertices_Number () {
	int max = 0;
	
	for (int i = 0; i < ct_number; i++)	{ 
		int nb = cluster_tree[i]->Get_Nb_Vertices ();
		if (max < nb)
			max = nb;
	}
	return max;
}


//-----------------------------------
// function computing cluster trees
//-----------------------------------


// makes a tree from the cluster list lc
void Tree_Decomposition::Make_Tree (pair<int,Light_List<Cluster_Node *> > lc) {  
	// we count the number of clusters in the list lc
	int nb_clusters = lc.first;

  // we compute the cluster tree by connecting, at each step, a new cluster from lc to a cluster chosen previously with which it shares the largest intersection (we use the Prim's algorithm)
  	
  // initialization of keys
  int * key;
  key = new int [nb_clusters];
  Cluster_Node ** parent;
  parent = new Cluster_Node * [nb_clusters];
  int * index ;
  index = new int [nb_clusters];
  Cluster_Node ** clusters;
  clusters = new Cluster_Node * [nb_clusters];

	Light_List<Cluster_Node *>::iterator iter_cl = lc.second.begin();
  for (int i = 0; i < nb_clusters; i++) {
    key[i] = 0;
    parent[i] = 0;
    index[i] = i;
    clusters[i] = *iter_cl;
    iter_cl++;
  }

	// we apply the Prim's Algorithm
	int ind_max = -1;
	int max;
	Light_List<pair<vertex,int> >::iterator neighbor_iter;
	Light_Vector<vertex>::iterator iter_v;
	int inter, j, choice;
	  
  while (nb_clusters > 0) {
    max = -1;
    for (j = 0; j < nb_clusters; j++)
      if (key[index[j]] > max) {
        max = key[index[j]];
        ind_max = j;
      }
      else if ((key[index[j]] == max) && (index[j] < index[ind_max]))	ind_max = j;

		choice = index[ind_max];
		
		// we update the cluster tree
		if (max == 0) {
			// we create a new cluster tree
			ct_number++;
			if ((unsigned) ct_number >= cluster_tree.get_capacity ())  cluster_tree.expand(50);
      
			cluster_tree[ct_number-1] = new Cluster_Tree(g);
				
			// we add the selected cluster to the current cluster tree
			cluster_tree[ct_number-1]->Add_Cluster (clusters[choice]);
		}
		else {
				// we add the selected cluster to the current cluster tree
				cluster_tree[ct_number-1]->Add_Cluster (clusters[choice]);
				
				// we link the selected cluster with its parent cluster
				cluster_tree[ct_number-1]->Add_Child_Node (parent[choice],clusters[choice]);
			}
				
		// we update the index data structure
		nb_clusters--;
    index[ind_max] = index[nb_clusters];
     
    // we update the key values
    for (j = 0; j < nb_clusters; j++) {
			// we compute the intersection between the last selected cluster and the unchosen clusters
      
      inter = 0;
      for (iter_v = clusters[choice]->begin(); iter_v != clusters[choice]->end(); iter_v++)
        if (clusters[index[j]]->Is_Element (*iter_v))  inter++;

      if (inter > key[index[j]]) {
        key[index[j]] = inter;
        parent[index[j]] = clusters[choice];
      }
    }
  }

  delete [] key;
  delete [] parent;
  delete [] index;
  delete [] clusters;
}
