/// \file 
/// \brief Definitions related to the Tree_Decomposition class

#ifndef TREE_DECOMPOSITION_H
#define TREE_DECOMPOSITION_H

#include <light_vector.h>
#include <light_list.h>
#include "graphical_structure.h"
#include "cluster_tree.h"
#include "first_cluster_heuristic.h"
#include "next_vertices_heuristic.h"

class Tree_Decomposition {                      /// This class allows to represent tree-decompositions of graphical structures (see [Robertson and Seymour, 1986])
	protected:
		Graphical_Structure * g;										///< the graphical structure to which the tree-decomposition refers
		Light_Vector<Cluster_Tree *> cluster_tree;	///< the cluster trees of the tree-decomposition (we consider a cluster tree per connected componed of the graphical structure g)
		int ct_number;															///< the number of cluster trees of the tree-decomposition
    
 		First_Cluster_Heuristic * first_cluster;    ///< the heuristic exploited for computing the first cluster
		Next_Vertices_Heuristic * next_vertices;    ///< the heuristic exploited for choosing the next vertices

		
	public:
		// constructors and destructor
		Tree_Decomposition (Graphical_Structure * gs);		      ///< constructs an empty tree-decomposition related to the graphical structure gs
    Tree_Decomposition (Graphical_Structure * gs, First_Cluster_Heuristic * fc, Next_Vertices_Heuristic * nv); ///< constructs a tree-decomposition related to the graphical structure gs by exploiting the heuristic fc for computing the first cluster and the heuristic nv for choosing the next vertex to consider
		virtual ~Tree_Decomposition ();													///< destructor
		
		
		// basic functions on cluster tree
		int Nb_Clusters ();							  ///< returns the number of clusters of the tree-decomposition
		int Get_Max_Label ();							///< returns the largest label of clusters in a cluster tree
		int Get_Max_Vertices_Number ();		///< returns the largest number of vertices in a the cluster tree
		int Get_Cluster_Tree_Number();		///< returns the number of cluster trees of the considered tree-decomposition
		Cluster_Tree * Get_Cluster_Tree (int i);					///< returns the i th cluster tree of the considered tree-decomposition
		Graphical_Structure * Get_Graphical_Structure ();	///< returns the graphical structure to which the tree-decomposition is related

		// function computing cluster trees
		virtual void Make_Tree (pair<int,Light_List<Cluster_Node *> > lc);		///< makes a tree from the cluster list lc
};


//-----------------------------
// inline function definitions
//-----------------------------

//-----------------
// basic functions 
//-----------------


// returns the number of cluster trees of the considered tree-decomposition
inline int Tree_Decomposition::Get_Cluster_Tree_Number() {
	return ct_number;
}


// returns the i th cluster tree of the considered tree-decomposition
inline Cluster_Tree * Tree_Decomposition::Get_Cluster_Tree (int i) {
	if (i < ct_number)	return cluster_tree[i];
	else return 0;
}


// returns the graphical structure to which the tree-decomposition is related
inline Graphical_Structure * Tree_Decomposition::Get_Graphical_Structure () {
	return g;
}

#endif
